@@ -1,5 +1,5 @@
 '''
-* Copyright 2015-2020 European Atomic Energy Community (EURATOM)
+* Copyright 2015-2025 European Atomic Energy Community (EURATOM)
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
   will be approved by the European Commission - subsequent
@@ -34,7 +34,7 @@ class CoordTransformer:
         self.set_transform_actions(transform_actions)
         self.x_pixels = orig_x
         self.y_pixels = orig_y
-        self.offset = offset
+        self.offset = (int(offset[0]),int(offset[1]))
         self.pixel_aspectratio = paspect
 
 
@@ -134,7 +134,7 @@ class CoordTransformer:
             x_offset (int) : Horizontal offset in pixels
             y_offset (int) : Vertical offset in pixels
         """
-        self.offset = (x_offset,y_offset)
+        self.offset = (int(x_offset),int(y_offset))
 
 
 
