@@ -1,5 +1,5 @@
 '''
-* Copyright 2015-2019 European Atomic Energy Community (EURATOM)
+* Copyright 2015-2025 European Atomic Energy Community (EURATOM)
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
   will be approved by the European Commission - subsequent
@@ -23,7 +23,7 @@
 '''
 Geometry matrix module for Calcam.
 
-Written by Scott Silburn & James Harrison.
+Written by Scott Silburn, James Harrison & Artur Perek.
 '''
 import multiprocessing
 import copy
@@ -35,18 +35,27 @@ import random
 import numpy as np
 import scipy.sparse
 import scipy.io
+import h5py
+
+# This is a workaround for Anaconda, where
+# the triangle package distribution appears to be broken.
+# Thanks Anaconda!
+try:
+    import triangle
+except:
+    trignale = None
 
 import matplotlib.pyplot as plt
-import matplotlib.cm as cm
 from matplotlib.patches import Polygon as PolyPatch
 from matplotlib.collections import PatchCollection
 import matplotlib.path as mplpath
 
+# The Matplotlib cmap API changed at v3.9.0
 try:
-    import meshpy.triangle
-    meshpy_err = None
-except Exception as e:
-    meshpy_err = '{:}'.format(e)
+    from matplotlib.colormaps import get_cmap
+except ImportError:
+    from matplotlib.cm import get_cmap
+
 
 from . import config
 from . import misc
@@ -243,9 +252,9 @@ class PoloidalVolumeGrid:
             
             # Sort the intersections by length along the sight-line.
             # Also round t_ray to 9 figures because we'll want to find unique values
-            # of it shortly, so round to something well over machine precision.
+            # of it shortly, so round to something slightly larger than the expected precision.
             sort_order = np.argsort(t_ray)
-            t_ray = t_ray[sort_order].round(decimals=12)
+            t_ray = t_ray[sort_order].round(decimals=9)
             seg_inds = seg_inds[sort_order]
             
             # This will be the output list of cell indices
@@ -287,7 +296,7 @@ class PoloidalVolumeGrid:
 
 
 
-    def plot(self,data=None,clim=None,cmap=None,line_colour=(0,0,0),cell_linewidth=None,cblabel='',axes=None):
+    def plot(self,data=None,clim=None,cmap=None,line_colour=(0,0,0),cell_linewidth=None,cblabel='',axes=None,cell_numbers=False):
         '''
         Either plot a given data vector on the grid, or if no data vector is given,
         plot the grid itself.
@@ -308,12 +317,14 @@ class PoloidalVolumeGrid:
                                               If set to None, the wall and cell boundaries are not drawn.
 
             cell_linewidth (float)          : Line width to use to show the grid cell boundaries. If set to 0, the grid \
-                                              cell boundaries will not be drawn.
+                                              cell boundaries will not be drawn. Default=0.25 if no data is given, or 0 if plotting data.
 
             cblabel (str)                   : Label for the data colour bar, if plotting data.
 
             axes (matplotlib.pyplot.Axes)   : Matplotlib axes on which to plot. If not given, a new figure will be created.
 
+            cell_numbers (bool)             : If set to True, displays the indices of the grid cells as text at the grid cell centre. \
+                                              This is itended mainly for debugging and can be very slow on some systems.
 
         Returns:
             
@@ -340,7 +351,7 @@ class PoloidalVolumeGrid:
             if data.size != self.n_cells:
                 raise ValueError('Data vector is the wrong length! Data has {:d} values but the grid gas {:d} cells.'.format(data.size,self.n_cells))
 
-            cmap = cm.get_cmap(cmap)
+            cmap = get_cmap(cmap)
 
             # If we have data, grid cell edges default to off
             if cell_linewidth is None:
@@ -372,6 +383,8 @@ class PoloidalVolumeGrid:
             pcoll.set_cmap(cmap)
             if clim is not None:
                 pcoll.set_clim(clim)
+        else:
+            pcoll.set_facecolor('None')
 
         pcoll.set_edgecolor(line_colour)
         pcoll.set_linewidth(cell_linewidth)
@@ -396,6 +409,14 @@ class PoloidalVolumeGrid:
         if data is not None:
             plt.colorbar(pcoll,label=cblabel if cblabel is not None else '')
 
+        # If requested, write the cell numbers
+        if cell_numbers:
+            for cell_ind in range(self.n_cells):
+                verts =  np.vstack( [ self.vertices[self.cells[cell_ind,i],:] for i in range(verts_per_cell)] )
+                r = verts[:,0].mean()
+                z = verts[:,1].mean()
+                plt.text(r,z,cell_ind,ha='center',va='center',fontsize=6,clip_on=True)
+
         # Prettification - set appropriate zoom for the grid extent.
         rmin,rmax,zmin,zmax = self.extent
         headroom = 0.1
@@ -577,6 +598,12 @@ class GeometryMatrix:
                                                the 2D image array in to the 1D data vector. \
                                                Default 'C' goes left-to-right, then row-by-row (NumPy default), \
                                                alternatively 'F' goes top-to-bottom, then column-by-column (MATLAB default).
+
+        trim_rows (bool)                     : Whether to automatically remove all-zero matrix rows, i.e. rows corresponding to pixels \
+                                               which don't see any of the cells in the inversion grid.
+
+        trim_columns (bool)                  : Whether to automatically remove all-zero matrix columns, i.e. columns corresponding \
+                                               to grid cells not seen by any pixel. This will also remove these cells from the grid object.
         
         calc_status_callback (callable)      : Callable which takes a single argument, which will be called with \
                                                status updates about the calculation. It will be called with either \
@@ -585,7 +612,7 @@ class GeometryMatrix:
                                                to stdout.  If set to None, no status updates are issued.
 
     '''
-    def __init__(self,grid,raydata,pixel_order='C',calc_status_callback = misc.LoopProgPrinter().update):
+    def __init__(self,grid,raydata,pixel_order='C',trim_rows=True,trim_columns=True,calc_status_callback = misc.LoopProgPrinter().update):
 
         if grid is not None and raydata is not None:
 
@@ -619,12 +646,11 @@ class GeometryMatrix:
             
             # Number of grid cells and sight lines
             n_cells = grid.n_cells
-            n_los = raydata.x.size
+            n_los = raydata.transform.x_pixels * raydata.transform.y_pixels
     
             # Flatten out the ray start and end coords
-            ray_start_coords = raydata.ray_start_coords.reshape(-1,3,order=self.pixel_order)
-            ray_end_coords = raydata.ray_end_coords.reshape(-1,3,order=self.pixel_order)
-
+            ray_start_coords = raydata.get_ray_start().reshape(-1,3,order=self.pixel_order)
+            ray_end_coords = raydata.get_ray_end().reshape(-1,3,order=self.pixel_order)
 
             # Multi-threadedly loop over each sight-line in raydata and calculate its matrix row.
             # Store the results as coords + data then build the matrix after, because that is much faster.
@@ -662,19 +688,20 @@ class GeometryMatrix:
             if calc_status_callback is not None:
                 calc_status_callback(1.)
             
-            
-            # Remove any grid cells + matrix columns which have no sight-line coverage.
-            unused_cells = np.where(np.abs(self.data.sum(axis=0)) == 0)[1]
-            self.grid.remove_cells(unused_cells)
+            if trim_columns:
+                # Remove any grid cells + matrix columns which have no sight-line coverage.
+                unused_cells = np.where(np.abs(self.data.sum(axis=0)) == 0)[1]
+                self.grid.remove_cells(unused_cells)
 
-            used_cols = np.where(self.data.sum(axis=0) > 0)[1]
-            self.data = self.data[:,used_cols]
+                used_cols = np.where(self.data.sum(axis=0) > 0)[1]
+                self.data = self.data[:,used_cols]
 
-            # Set the pixel mask to exclude pixels which do not contribute to any grid cells and remove corresponding rows.
-            if self.pixel_mask is not None:
-                used_pixels = np.squeeze(np.array(np.abs(self.data.sum(axis=1))) > 1e-14)
-                self.pixel_mask = used_pixels.reshape(imdims,order=self.pixel_order)
-                self.data = self.data[used_pixels,:]
+            if trim_rows:
+                # Set the pixel mask to exclude pixels which do not contribute to any grid cells and remove corresponding rows.
+                if self.pixel_mask is not None:
+                    used_pixels = np.squeeze(np.array(np.abs(self.data.sum(axis=1))) > 1e-14)
+                    self.pixel_mask = used_pixels.reshape(imdims,order=self.pixel_order)
+                    self.data = self.data[used_pixels,:]
 
 
 
@@ -1034,7 +1061,6 @@ class GeometryMatrix:
         
         # Loop over each intersection
         for i in range(positions.size):
-            
             if len(in_cell) == 0:
                 # Entering the grid
                 in_cell = intersected_cells[i]
@@ -1042,7 +1068,6 @@ class GeometryMatrix:
             else:
                 # Going from one cell to another         
                 leaving_cell = list(intersected_cells[i] & in_cell)
-                
                 if len(leaving_cell) == 1:
                     
                     #row_out[0,leaving_cell[0]] = row_out[0,leaving_cell[0]] + (positions[i] - positions[i-1])
@@ -1051,16 +1076,39 @@ class GeometryMatrix:
                     in_cell.remove(leaving_cell[0])
                 
                 else:
-                    '''
-                    # Make a plot to show where on the grid the error has occured.
-                    self.grid.plot()
-                    self.grid.get_cell_intersections(ray_start_coords,ray_end_coords,plot=True)
-                    problem_point = ray_start_coords + positions[i]*(ray_end_coords - ray_start_coords)/ray_length
-                    plt.plot(np.sqrt(problem_point[0]**2 + problem_point[1]**2),problem_point[2],'bo')
-                    plt.title('Error for: {:},{:}'.format(ray_start_coords,ray_end_coords))
-                    plt.show()
-                    '''
-                    raise Exception('Error generating geometry matrix row: could not identify which grid cell the LoS left.')
+                    recovered_from_error = False
+                    if i < positions.size - 1:
+                        # Due to maths accuracy we can end up missing an intersection very close to a vertex sometimes.
+                        # In this case we assume we only missed a trivial distance and we're now in whatever cell we see ourselves
+                        # leaving next
+                        leaving_cell = list(in_cell)
+                        data[leaving_cell[0]] = data[leaving_cell[0]] + (positions[i] - positions[i - 1])
+                        in_cell = intersected_cells[i] & intersected_cells[i+1]
+                        if len(in_cell) == 1:
+                            recovered_from_error = True
+                    else:
+                        # If we find ourselves leaving the grid without exiting the last cell we were in, simply assume we exit that cell here
+                        # (this can happen for grids with very tiny cells at the edges)
+                        leaving_cell = list(in_cell)
+                        data[leaving_cell[0]] = data[leaving_cell[0]] + (positions[i] - positions[i - 1])
+                        in_cell = set()
+                        recovered_from_error = True
+
+                    if not recovered_from_error:
+                        try:
+                            # Make a plot to show where on the grid the error has occured.
+                            self.grid.plot(cell_numbers=True)
+                            self.grid.get_cell_intersections(ray_start_coords,ray_end_coords,plot=True)
+                            problem_point = ray_start_coords + positions[i]*(ray_end_coords - ray_start_coords)/ray_length
+                            plt.plot(np.sqrt(problem_point[0]**2 + problem_point[1]**2),problem_point[2],'bo')
+                            plt.title('This plot is here because an error has occured\nIt may be helpful to zoom in to the blue point and save a figure to report this error\nOnce this plot window is closed, an exception will be raised in Calcam.\nLoS start, end causing error: {:},{:}'.format(ray_start_coords,ray_end_coords))
+                            plt.xlim([np.sqrt(problem_point[0]**2 + problem_point[1]**2) - 0.05,np.sqrt(problem_point[0]**2 + problem_point[1]**2) + 0.05])
+                            plt.ylim([problem_point[2]-0.05,problem_point[2]+0.05])
+                            plt.show()
+                        except:
+                            pass
+
+                        raise Exception('Error generating geometry matrix row: could not identify which grid cell the LoS left.')
         
         
         # If the sight line ends inside a cell, add the length it was inside that cell.
@@ -1139,7 +1187,7 @@ class GeometryMatrix:
                            'grid_history':self.history['grid'],
                            'pixel_mask':self.pixel_mask,
                            'grid_type':self.grid.__class__.__name__,
-                           'im_transforms': np.array(self.image_geometry.transform_actions,dtype=np.object),
+                           'im_transforms': np.array(self.image_geometry.transform_actions,dtype=object),
                            'im_px_aspect':self.image_geometry.pixel_aspectratio,
                            'im_coords':self.image_coords
                          }
@@ -1150,24 +1198,92 @@ class GeometryMatrix:
         '''
         Load geometry matrix from a MATLAB file.
         '''
-        f = scipy.io.loadmat(filename)
-        
-        
-        
-        self.binning = float(f['binning'][0,0])
-        self.pixel_order = str(f['pixel_order'][0])
-        self.pixel_mask = f['pixel_mask']
-        self.history = {'los':str(f['sightline_history'][0]),'grid':str(f['grid_history'][0]),'matrix':str(f['matrix_history'][0])}
-        self.image_coords = str(f['im_coords'][0])
+        try:
+            # Try to load with scipy for older .mat formats
+            f = scipy.io.loadmat(filename)
+            is_v73 = False
+        except Exception:
+            # If scipy fails, try with h5py for v7.3 .mat files
+            try:
+                f = h5py.File(filename, 'r')
+                is_v73 = True
+            except Exception as e:
+                # Handle failure to open file with both methods
+                raise Exception("Failed to load the file '{:s}' with either scipy or h5py. Error: {:}".format(filename,e))
 
-        self.grid = PoloidalVolumeGrid(f['grid_verts'],f['grid_cells'],f['grid_wall'],src=self.history['grid'])
-        self.data = f['geom_mat'].tocsr()
-        
-        self.image_geometry = CoordTransformer()
-        self.image_geometry.set_transform_actions(f['im_transforms'])
-        self.image_geometry.set_pixel_aspect(f['im_px_aspect'],relative_to='Original')
-        self.image_geometry.set_image_shape(*self.binning*np.array(self.pixel_mask.shape[::-1]),coords=self.image_coords)
+        if is_v73:
+            # Load scalar values
+            self.binning = float(f['binning'][0][0])
+            self.image_coords = f['im_coords'][()].tobytes().decode('utf-16')
 
+            self.image_geometry = CoordTransformer()  
+
+            image_pixel_aspect_data = np.array(f['im_px_aspect'])
+
+            # Use the extracted data to set the pixel aspect in the image geometry
+            self.image_geometry.set_pixel_aspect(image_pixel_aspect_data[0, 0], relative_to='Original')
+
+            # Print the numerical data in the numpy array format
+
+            self.pixel_mask = f['pixel_mask'][:].T
+            self.pixel_mask = np.array(self.pixel_mask, dtype=bool)
+ 
+            # Load 2D arrays and transpose them
+            grid_verts_transposed = f['grid_verts'][:].T
+            grid_cells_transposed = f['grid_cells'][:].T
+            grid_wall_transposed = f['grid_wall'][:].T
+ 
+            # Convert numerical arrays to ASCII strings
+            self.pixel_order = ''.join(chr(int(i)) for i in f['pixel_order'][:, 0])
+ 
+            # Load historical data
+            self.history = {
+                'los': ''.join(chr(int(i)) for i in f['sightline_history'][:, 0]),
+                'grid': ''.join(chr(int(i)) for i in f['grid_history'][:, 0]),
+                'matrix': ''.join(chr(int(i)) for i in f['matrix_history'][:, 0])
+            }
+ 
+            # Initialize PoloidalVolumeGrid with transposed arrays
+            self.grid = PoloidalVolumeGrid(grid_verts_transposed, grid_cells_transposed, grid_wall_transposed, src=self.history['grid'])
+
+            # Load sparse matrix data
+            data = f['geom_mat/data'][:]
+            ir = f['geom_mat/ir'][:]
+            jc = f['geom_mat/jc'][:]
+
+            # Check and adjust the shape of the sparse matrix
+            expected_rows = np.prod(self.pixel_mask.shape)
+            self.data = scipy.sparse.csr_matrix((data, ir, jc))
+            self.data = self.data.T
+
+            if self.data.shape[0] < expected_rows:
+                rows_to_add = expected_rows - self.data.shape[0]
+                zeros_matrix = scipy.sparse.csr_matrix((rows_to_add, self.data.shape[1]))
+                self.data = scipy.sparse.vstack([self.data, zeros_matrix])
+
+            # Set image geometry
+            self.image_geometry.set_image_shape(*self.binning * np.array(self.pixel_mask.shape[::-1]), coords=self.image_coords)
+
+            # Close the HDF5 file
+            f.close()
+
+        else:    
+        # Processing for older .mat formats
+            self.binning = float(f['binning'][0,0])
+            self.pixel_order = str(f['pixel_order'][0])
+            self.pixel_mask = f['pixel_mask']
+            self.history = {
+                'los': str(f['sightline_history'][0]),
+                'grid': str(f['grid_history'][0]),
+                'matrix': str(f['matrix_history'][0])
+            }
+            self.image_coords = str(f['im_coords'][0])
+            self.grid = PoloidalVolumeGrid(f['grid_verts'], f['grid_cells'], f['grid_wall'], src=self.history['grid'])
+            self.data = f['geom_mat'].tocsr()
+            self.image_geometry = CoordTransformer()  # Ensure CoordTransformer is defined/imported
+            self.image_geometry.set_transform_actions(f['im_transforms'])
+            self.image_geometry.set_pixel_aspect(f['im_px_aspect'], relative_to='Original')
+            self.image_geometry.set_image_shape(*self.binning * np.array(self.pixel_mask.shape[::-1]), coords=self.image_coords)
 
     def _save_txt(self,filename):
         '''
@@ -1247,9 +1363,9 @@ class GeometryMatrix:
         msg = msg + self.history['los'] + '\n'
 
         if self.image_coords.lower() == 'original':
-        	im_shape = self.image_geometry.get_original_shape()
+            im_shape = self.image_geometry.get_original_shape()
         elif self.image_coords.lower() == 'display':
-        	im_shape = self.image_geometry.get_display_shape()
+            im_shape = self.image_geometry.get_display_shape()
 
         msg = msg + 'Image dimensions:  {:d} x {:d} px\n'.format(im_shape[0],im_shape[1])
         msg = msg + 'Image orientation: {:s}\n'.format(self.image_coords)
@@ -1318,7 +1434,7 @@ def squaregrid(wall_contour,cell_size,rmin=None,rmax=None,zmin=None,zmax=None):
                                            
     Returns:
         
-        calcam.PoloidalVolumeGrid    : Generated grid.
+        calcam.gm.PoloidalVolumeGrid    : Generated grid.
         
     '''
     
@@ -1400,12 +1516,10 @@ def squaregrid(wall_contour,cell_size,rmin=None,rmax=None,zmin=None,zmax=None):
 
 
 
-def trigrid(wall_contour,max_cell_scale,rmin=None,rmax=None,zmin=None,zmax=None,**kwargs):
+def trigrid(wall_contour,max_cell_scale,rmin=None,rmax=None,zmin=None,zmax=None,triopts=[]):
     '''
     Create a reconstruction grid with triangular grid cells conforming to the 
-    wall contour. Requires the MeshPy package (https://pypi.org/project/MeshPy/) 
-    to be installed, since it uses J. Shewchuk's "triangle" via the MeshPy to 
-    generate the grid.
+    wall contour, using Jonathan Richard Shewchuk’s "triangle" triangulation library.
     
     Parameters:
         
@@ -1419,19 +1533,18 @@ def trigrid(wall_contour,max_cell_scale,rmin=None,rmax=None,zmin=None,zmax=None,
         floats rmin, rmax, zmin, zmax       : Optional limits of the grid extent in the R, Z plane. \
                                               Any combination of these may or may not be given; if none \
                                               are given the entire wall contour interior is gridded.
-                                           
-    Any additional keyword arguments will be passed directly to the triangle mesher, meshpy.triangle.build().
-    This can be used to further control the meshing; see the MeshPy documentation for available arguments.
-    
-                                           
+
+        trioppts (list of str)              : List of string options to pass to triangle.triangulate(), see triangle \
+                                              `trignale documentation page <https://rufat.be/triangle/API.html>`_ for possible options.
+
     Returns:
         
-        calcam.PoloidalVolumeGrid    : Generated grid.
+        calcam.gm.PoloidalVolumeGrid    : Generated grid.
         
     '''
-    # Before trying to generate a triangular grid, check we have the meshpy package available.
-    if meshpy_err is not None:
-        raise Exception('This function requires the MeshPy module (https://pypi.org/project/MeshPy), which could not be imported: {:s}'.format(meshpy_err))
+
+    if triangle is None:
+        raise ImportError('"triangle" Python package not available; cannot create triangular grid. Please try installing triangle using pip then try again.')
 
     # If given a machine name for the wall contour, get the R,Z contour
     if type(wall_contour) is str:
@@ -1506,29 +1619,112 @@ def trigrid(wall_contour,max_cell_scale,rmin=None,rmax=None,zmin=None,zmax=None,
 
         geometry = {'vertices': verts, 'segments' : segments}
 
+    opts = 'pa{:.6f}'.format(max_cell_area)
 
-    mesher_kwargs = dict(kwargs)
-    mesher_kwargs['max_volume'] = max_cell_area
+    for user_opt in triopts:
+        if not (user_opt.startswith('p') or user_opt.startswith('a')):
+            opts = opts + user_opt
 
-    # Run triangle to make the grid!
-    # Do this in a separate process so that if the triangle library has a fail,
-    # it doesn't drag this whole python process down with it.
-    q = multiprocessing.Queue()
-    triprocess = multiprocessing.Process(target=_run_triangle_python,args=(geometry,q),kwargs=mesher_kwargs)
-    triprocess.start()
-    while triprocess.is_alive() and q.empty():
-        time.sleep(0.05)
-    
-    if not q.empty():
-        mesh = q.get()
-        if isinstance(mesh,Exception):
-            raise mesh
+    mesh = triangle.triangulate(geometry,opts)
+
+    return PoloidalVolumeGrid(mesh['vertices'],mesh['triangles'],wall_contour,src='Wall-conforming triangular mesh with ~{:.1f}cm cells generated with trigrid()'.format(max_cell_scale*1e2))
+
+
+def solps_grid(solps_file,wall_contour,rmin=None,rmax=None,zmin=None,zmax=None,cell_attrs=False):
+    """
+    Load a grid from a SOLPS-ITER DivGeo-Carre file.
+
+    Parameters:
+
+        solps_file (str)                    : Filename of DivGeo-Carre .geo file containing grid coordinates to load.
+
+        wall_contour (str or numpy.ndarray) : Either the name of a Calcam CAD model \
+                                              from which to use the wall contour, or an \
+                                              N x 2 array of R,Z points defining the machine wall.
+
+        rmin, rmax, zmin, zmax (float)      : Optional limits of the grid extent in the R, Z plane. \
+                                              Any combination of these may or may not be given; if none \
+                                              are given the entire SOLPS-ITER grid is loaded.
+
+        cell_attrs (bool)                   : Whether to return the cell centre coordinates and cell \
+                                              indices as provided by SOLPS-ITER.
+
+    Returns:
+
+        calcam.gm.PoloidalVolumeGrid    : The reconstruction grid.
+
+        if cell_attrs is set to true, also returns an Nx8 NumPy array where N is the number of \
+        grid cells. The 8 columns of this array contain for each cell: Poloidal index, Radial index, R centre,\
+        Z centre, and indices in to the resulting grid object's vertex list of the 4 vertices of the cell.
+
+    """
+
+    # If given a machine name for the wall contour, get the R,Z contour
+    if type(wall_contour) is str:
+        wall_contour = _get_ccm_wall(wall_contour)
+
+    # If we have a wall contour which joins back up with itself, remove
+    # the last point.
+    if np.abs(wall_contour[0,:] - wall_contour[-1,:]).max() < 1e-15:
+        wall_contour = wall_contour[:-1]
+
+    vertices = []
+    cells = []
+    centres = []
+    pol_inds = []
+    rad_inds = []
+
+    with open(solps_file,'r') as f:
+
+        npol,nrad = [int(n)+2 for n in f.readline().split()]
+
+        ncells = npol*nrad
+
+        for index in range(ncells):
+
+            row = np.array([float(d) for d in f.readline().split()])
+
+            pol,rad = np.array(row[0:2])  # Poloidal and radial index on SOLPS grid
+            R,Z = row[4:12:2], row[5:13:2]
+            cell_centre = row[2:4]
+
+            if not pol in pol_inds and not rad in rad_inds:
+                if rmin and cell_centre[0] < rmin:
+                    continue
+                if rmax and cell_centre[0] > rmax:
+                    continue
+                if zmin and cell_centre[1] < zmin:
+                    continue
+                if zmax and cell_centre[1] > zmax:
+                    continue
+
+            pol_inds.append(pol)
+            rad_inds.append(rad)
+
+            verts = list(zip(R,Z))
+            indices = []
+            for vert in verts:
+                if vert in vertices:
+                    indices.append(vertices.index(vert))
+                else:
+                    vertices.append(vert)
+                    indices.append(vertices.index(vert))
+            cells.append(indices)
+            centres.append(cell_centre)
+
+    vertices = np.array(vertices)
+    cell_info = np.column_stack((pol_inds,rad_inds, centres, cells))
+    cell_info = cell_info[cell_info[:,0].argsort()]
+    cell_info = cell_info[cell_info[:,1].argsort(kind='mergesort')]
+
+    cell_info[:, [-2,-1]] = cell_info[:,[-1,-2]]  # Swap coords order to match Calcam convention
+
+    grid = PoloidalVolumeGrid(vertices, cell_info[:, -4:].astype(int), wall_contour,src="Loaded from SOLPS-ITER DivGeo-Carre file: {:s}".format(solps_file))
+
+    if cell_attrs:
+        return grid,cell_info
     else:
-        raise Exception('Call to meshpy.triangle to generate the grid failed silently. The triangle library can be a bit flaky; maybe try again and/or change settings.')
-
-    return PoloidalVolumeGrid(mesh['vertices'],mesh['cells'],wall_contour,src='Wall-conforming triangular mesh with ~{:.1f}cm cells generated with trigrid()'.format(max_cell_scale*1e2))
-
-
+        return grid
 
 
 def _get_ccm_wall(model_name):
