@@ -1,5 +1,5 @@
 '''
-* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
+* Copyright 2015-2025 European Atomic Energy Community (EURATOM)
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
   will be approved by the European Commission - subsequent
@@ -139,6 +139,7 @@ class AlignmentCalib(CalcamGUIWindow):
         self.intrinsics_calib = None
 
         self.extrinsics_src = None
+        self.chessboard_pointpairs = None
 
         self.calibration = None
 
@@ -151,7 +152,7 @@ class AlignmentCalib(CalcamGUIWindow):
             self.edge_detect.setToolTip('Requires OpenCV 2.4.6 or newer; you have {:s}'.format(cv2.__version__))
 
         self.fov_enabled = False
-        self.viewdir_at_cc = True
+        self.viewdir_at = 'cc'
 
         # Start the GUI!
         self.show()
