@@ -1,5 +1,5 @@
 '''
-* Copyright 2015-2021 European Atomic Energy Community (EURATOM)
+* Copyright 2015-2025 European Atomic Energy Community (EURATOM)
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
   will be approved by the European Commission - subsequent
@@ -90,6 +90,16 @@ class CADEdit(CalcamGUIWindow):
         self.replace_file_button.clicked.connect(self.edit_feature)
         self.mesh_scale_box.valueChanged.connect(self.edit_feature)
         self.del_feature_button.clicked.connect(self.edit_feature)
+        self.meshup_box.currentIndexChanged.connect(self.edit_feature)
+        self.z_rotate_box.valueChanged.connect(self.edit_feature)
+        self.z_rotate_box.valueChanged.connect(self.edit_feature)
+        self.handedness_box.currentIndexChanged.connect(self.edit_feature)
+        self.slice_apply_button.clicked.connect(self.update_xsection)
+        self.proj_perspective.toggled.connect(self.set_projection)
+        self.no_slicing_rb.clicked.connect(self.update_slicing_options)
+        self.cakeslice_rb.clicked.connect(self.update_slicing_options)
+        self.chordslice_rb.clicked.connect(self.update_slicing_options)
+        self.zslice_checkbox.toggled.connect(self.update_slicing_options)
 
         self.new_group_button.clicked.connect(self.add_group)
 
@@ -122,6 +132,12 @@ class CADEdit(CalcamGUIWindow):
             self.reset()
             
 
+    def on_model_load(self):
+        model_extent = self.cadmodel.get_extent()
+        rmax = np.max(np.abs(model_extent[:4]))
+        self.chordslice_r.setValue(rmax/2)
+        self.zslice_zmin.setValue(model_extent[4]-1e-3)
+        self.zslice_zmax.setValue(model_extent[5]+1e-3)
 
 
     def on_featuretree_change(self,item):
@@ -196,7 +212,7 @@ class CADEdit(CalcamGUIWindow):
         namedialog = NameInputDialog(self,'Add new part group','Enter a name for the new part group:')
         namedialog.exec()
 
-        if namedialog.result() == 1:
+        if namedialog.result() == namedialog.Accepted:
 
             name = str(namedialog.text_input.text())
 
@@ -204,6 +220,15 @@ class CADEdit(CalcamGUIWindow):
             self.extra_groups[self.cadmodel.model_variant].append(name)
             self.update_feature_tree()
 
+    def set_projection(self):
+
+        if self.proj_perspective.isChecked():
+            self.interactor3d.set_projection('perspective')
+        elif self.proj_orthographic.isChecked():
+            self.interactor3d.set_projection('orthographic')
+
+        self.update_viewport_info()
+        self.refresh_3d()
 
     def load_wall_contour(self):
 
@@ -246,7 +271,7 @@ class CADEdit(CalcamGUIWindow):
 
         dialog = AreYouSureDialog('Remove wall contour?','Are you sure you want to remove the wall contour?')
         dialog.exec()
-        if dialog.result() == 1:
+        if dialog.result() == dialog.Accepted:
             self.show_contour_checkbox.setChecked(False)
             self.show_contour_checkbox.setEnabled(False)
             self.wall_contour = None
@@ -257,34 +282,7 @@ class CADEdit(CalcamGUIWindow):
 
     def edit_feature(self):
 
-        if self.sender() is self.replace_file_button:
-            mesh_path = self.browse_for_mesh()
-
-            if mesh_path is not None:
-                self.cadmodel.set_features_enabled(False,self.selected_feature)
-                if self.cadmodel.def_file is not None:
-                    old_path = self.model_features[self.cadmodel.model_variant][self.selected_feature]['mesh_file']
-                    if self.cadmodel.def_file.get_temp_path() in old_path:
-                        self.removed_mesh_files.append(old_path.replace(self.cadmodel.def_file.get_temp_path() + os.sep,''))
-
-                self.model_features[self.cadmodel.model_variant][self.selected_feature]['mesh_file'] = mesh_path
-                self.cadmodel.features[self.selected_feature].filename = mesh_path
-                self.cadmodel.features[self.selected_feature].filetype = mesh_path.split('.')[-1].lower()
-                self.cadmodel.features[self.selected_feature].polydata = None
-                self.cadmodel.features[self.selected_feature].solid_actor = None
-                self.cadmodel.features[self.selected_feature].edge_actor = None
-                self.cadmodel.set_features_enabled(True,self.selected_feature)
-
-        elif self.sender() is self.mesh_scale_box:
-            self.cadmodel.set_features_enabled(False,self.selected_feature)
-            self.model_features[self.cadmodel.model_variant][self.selected_feature]['mesh_scale'] = self.mesh_scale_box.value()
-            self.cadmodel.features[self.selected_feature].scale = self.mesh_scale_box.value()
-            self.cadmodel.features[self.selected_feature].polydata = None
-            self.cadmodel.features[self.selected_feature].solid_actor = None
-            self.cadmodel.features[self.selected_feature].edge_actor = None
-            self.cadmodel.set_features_enabled(True,self.selected_feature)
-
-        elif self.sender() is self.cad_colour_choose_button:
+        if self.sender() is self.cad_colour_choose_button:
 
             self.set_cad_colour()
             self.model_features[self.cadmodel.model_variant][self.selected_feature]['colour'] = self.cadmodel.features[self.selected_feature].colour
@@ -292,7 +290,7 @@ class CADEdit(CalcamGUIWindow):
         elif self.sender() is self.del_feature_button:
             dialog = AreYouSureDialog(self,'Confirm Delete Model Part','Are you sure you want to remove the model part "{:s}"?'.format(self.selected_feature.split('/')[-1]))
             dialog.exec()
-            if dialog.result():
+            if dialog.result() == dialog.Accepted:
 
                 # Make sure any dragging & dropping is sorted.
                 self.update_feature_tree()
@@ -306,7 +304,51 @@ class CADEdit(CalcamGUIWindow):
                 if self.current_group is not None:
                     self.cadmodel.groups[self.current_group].remove(self.selected_feature)
                 self.update_feature_tree()
-    
+
+        else:
+            if self.sender() is self.replace_file_button:
+                mesh_path = self.browse_for_mesh()
+
+                if mesh_path is not None:
+                    self.cadmodel.set_features_enabled(False,self.selected_feature)
+                    if self.cadmodel.def_file is not None:
+                        old_path = self.model_features[self.cadmodel.model_variant][self.selected_feature]['mesh_file']
+                        if self.cadmodel.def_file.get_temp_path() in old_path:
+                            self.removed_mesh_files.append(old_path.replace(self.cadmodel.def_file.get_temp_path() + os.sep,''))
+
+                    self.model_features[self.cadmodel.model_variant][self.selected_feature]['mesh_file'] = mesh_path
+                    self.cadmodel.features[self.selected_feature].filename = mesh_path
+                    self.cadmodel.features[self.selected_feature].filetype = mesh_path.split('.')[-1].lower()
+
+            elif self.sender() is self.mesh_scale_box:
+                self.cadmodel.set_features_enabled(False,self.selected_feature)
+                self.model_features[self.cadmodel.model_variant][self.selected_feature]['mesh_scale'] = self.mesh_scale_box.value()
+                self.cadmodel.features[self.selected_feature].scale = self.mesh_scale_box.value()
+
+
+            elif self.sender() is self.meshup_box:
+                self.cadmodel.set_features_enabled(False,self.selected_feature)
+                self.model_features[self.cadmodel.model_variant][self.selected_feature]['mesh_up_direction'] = self.meshup_box.currentText()
+                self.cadmodel.features[self.selected_feature].mesh_up = self.meshup_box.currentText()
+
+
+            elif self.sender() is self.z_rotate_box:
+                self.cadmodel.set_features_enabled(False,self.selected_feature)
+                self.model_features[self.cadmodel.model_variant][self.selected_feature]['rotate_toroidal'] = self.z_rotate_box.value()
+                self.cadmodel.features[self.selected_feature].toroidal_rotation = self.z_rotate_box.value()
+
+
+            elif self.sender() is self.handedness_box:
+                self.cadmodel.set_features_enabled(False,self.selected_feature)
+                self.model_features[self.cadmodel.model_variant][self.selected_feature]['coord_handedness'] = 'right' if self.handedness_box.currentIndex() == 0 else 'left'
+                self.cadmodel.features[self.selected_feature].coord_handedness = 'right' if self.handedness_box.currentIndex() == 0 else 'left'
+
+            self.cadmodel.features[self.selected_feature].polydata = None
+            self.cadmodel.features[self.selected_feature].solid_actor = None
+            self.cadmodel.features[self.selected_feature].edge_actor = None
+            self.cadmodel.set_features_enabled(True, self.selected_feature)
+
+
         self.update_current_feature()
         self.unsaved_changes = True
         self.refresh_3d()
@@ -318,7 +360,7 @@ class CADEdit(CalcamGUIWindow):
             dialog = NameInputDialog(self,'Add Model Variant','Enter a name for the new model variant:')
 
             dialog.exec()
-            if dialog.result() == 1:
+            if dialog.result() == dialog.Accepted:
                 variant_name = str(dialog.text_input.text())
             else:
                 return
@@ -348,7 +390,7 @@ class CADEdit(CalcamGUIWindow):
         dialog = AreYouSureDialog(self,'Remove model variant','Are you sure you want to remove the model variant "{:s}"?'.format(self.cadmodel.model_variant))
         dialog.exec()
         
-        if dialog.result() == 1:
+        if dialog.result() == dialog.Accepted:
 
             variant_to_remove = self.cadmodel.model_variant
             index_to_remove = self.model_variant.currentIndex()
@@ -379,6 +421,7 @@ class CADEdit(CalcamGUIWindow):
         try:
             feature = self.cadmodel.features[self.selected_feature]
             self.component_settings.setEnabled(True)
+            self.coords_settings.setEnabled(True)
             self.del_feature_button.setEnabled(True)
             self.cad_colour_choose_button.setEnabled(True)
         except KeyError:
@@ -395,6 +438,7 @@ class CADEdit(CalcamGUIWindow):
 
         filename = feature.filename
         scale = feature.scale
+        upax = feature.mesh_up
 
         filesize = os.path.getsize(filename) / 1024.**2
 
@@ -407,6 +451,15 @@ class CADEdit(CalcamGUIWindow):
         self.mesh_scale_box.blockSignals(True)
         self.mesh_scale_box.setValue(scale)
         self.mesh_scale_box.blockSignals(False)
+        self.meshup_box.blockSignals(True)
+        self.meshup_box.setCurrentText(upax)
+        self.meshup_box.blockSignals(False)
+        self.handedness_box.blockSignals(True)
+        self.handedness_box.setCurrentIndex(0 if feature.coord_handedness == 'right' else 1)
+        self.handedness_box.blockSignals(False)
+        self.z_rotate_box.blockSignals(True)
+        self.z_rotate_box.setValue(feature.toroidal_rotation)
+        self.z_rotate_box.blockSignals(False)
 
         feature_extent = feature.get_polydata().GetBounds()
 
@@ -431,7 +484,7 @@ class CADEdit(CalcamGUIWindow):
         dialog = NameInputDialog(self,'Rename Model Variant','Enter a new name for the "{:s}" model variant:'.format(self.cadmodel.model_variant),init_text=self.cadmodel.model_variant)
 
         dialog.exec()
-        if dialog.result() == 1:
+        if dialog.result() == dialog.Accepted:
             new_name = str(dialog.text_input.text())
         else:
             return
@@ -612,7 +665,7 @@ class CADEdit(CalcamGUIWindow):
             filedialog.setNameFilter('Calcam CAD model definitions (*.ccm)')
             filedialog.exec()
 
-            if filedialog.result() == 1:
+            if filedialog.result() == dialog.Accepted:
                 filename = filedialog.selectedFiles()[0]
             else:
                 return
@@ -715,6 +768,7 @@ class CADEdit(CalcamGUIWindow):
             self.action_save.setEnabled(True)
 
         self.unsaved_changes = False
+        self.on_model_load()
         self.app.restoreOverrideCursor()
 
 
@@ -734,7 +788,7 @@ class CADEdit(CalcamGUIWindow):
         selected_view = str(self.viewlist.selectedItems()[0].text(0)).replace('*','')
         dialog = AreYouSureDialog(self,'Remove preset view','Are you sure you want to remove the view "{:s}"?'.format(selected_view))
         dialog.exec()
-        if dialog.result() == 1:
+        if dialog.result() == dialog.Accepted:
             self.cadmodel.views.pop(selected_view)
             if self.cadmodel.initial_view == selected_view:
                 self.cadmodel.initial_view = None
@@ -870,7 +924,14 @@ class CADEdit(CalcamGUIWindow):
                 init_name = '{:s}/{:s}'.format(self.current_group,init_name)
                 self.cadmodel.groups[self.current_group].append(init_name)
 
-            feature_dict = {'mesh_file':mesh_path,'default_enable':False,'mesh_scale':self.mesh_scale_box.value(),'colour':(0.75,0.75,0.75)}
+            feature_dict = {'mesh_file':mesh_path,'default_enable':False,'mesh_scale':self.mesh_scale_box.value(),'colour':(0.75,0.75,0.75),'mesh_up_direction':self.meshup_box.currentText(),'rotate_toroidal':self.z_rotate_box.value(),'coord_handedness':'right' if self.handedness_box.currentIndex() == 0 else 'left'}
+
+            # If adding a duplicate, append number to the end of the feature name to create unique names
+            if init_name in self.model_features[self.cadmodel.model_variant]:
+                n = 1
+                while init_name + ' ({:d})'.format(n) in self.model_features[self.cadmodel.model_variant]:
+                    n += 1
+                init_name = init_name + ' ({:d})'.format(n)
 
             self.model_features[self.cadmodel.model_variant][init_name] = copy.copy(feature_dict)
             self.cadmodel.features[init_name] = ModelFeature(self.cadmodel,feature_dict,abs_path=True)
@@ -1075,7 +1136,7 @@ class CADEdit(CalcamGUIWindow):
         filedialog.setNameFilter('Supported mesh files (*.stl *.obj)')
         filedialog.exec()
 
-        if filedialog.result() == 1:
+        if filedialog.result() == filedialog.Accepted:
             mesh_paths = [str(path).replace(os.sep,'/') for path in filedialog.selectedFiles()]
             if multiple:
                 return mesh_paths
@@ -1140,7 +1201,7 @@ class CADEdit(CalcamGUIWindow):
             filedialog.setNameFilter('Calcam CAD model definitions (*.ccm)')
             filedialog.exec()
 
-            if filedialog.result() == 1:
+            if filedialog.result() == filedialog.Accepted:
                 filename = filedialog.selectedFiles()[0]
             else:
                 return
@@ -1239,6 +1300,6 @@ class CADEdit(CalcamGUIWindow):
         if add_path_prompt:
             dialog = AreYouSureDialog(self,'Add to calcam configuration?',"Model saved.<br><br>The folder it is saved in:<br><br> {:s}<br><br>is currently not in Calcam's cad model search path,<br>so this model will not yet be detected by Calcam.<br><br>Would you like to add this path to Calcam's<br>model search path configuration now?".format(add_path_prompt))
             dialog.exec()
-            if dialog.result() == 1:
+            if dialog.result() == dialog.Accepted:
                 self.config.cad_def_paths.append(add_path_prompt)
                 self.config.save()
