@@ -1,5 +1,5 @@
 '''
-* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
+* Copyright 2015-2025 European Atomic Energy Community (EURATOM)
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
   will be approved by the European Commission - subsequent
@@ -22,7 +22,11 @@
 
 import copy
 
-import matplotlib.cm
+# The Matplotlib cmap API changed at v3.9.0
+try:
+    from matplotlib.colormaps import get_cmap
+except ImportError:
+    from matplotlib.cm import get_cmap
 
 from .core import *
 from .vtkinteractorstyles import CalcamInteractorStyle2D, CalcamInteractorStyle3D
@@ -38,7 +42,7 @@ type_description = {'alignment': 'Manual Alignment', 'fit':'Point pair fitting',
 cmaps = []
 for cmap in [('Grayscale','gray'),('Inferno','inferno'),('Magma','magma'),('Plasma','plasma'),('Viridis','viridis'),('JET','jet'),('HSV','hsv')]:
     try:
-        cmaps.append((cmap[0],matplotlib.cm.get_cmap(cmap[1])))
+        cmaps.append((cmap[0],get_cmap(cmap[1])))
     except Exception:
         continue
 
@@ -123,7 +127,7 @@ class ImageAnalyser(CalcamGUIWindow):
         self.movement_correction_button.clicked.connect(self.update_movement_correction)
         self.im_mapping_checkbox.toggled.connect(self.update_mapped_image)
         self.mapped_im_opacity_slider.valueChanged.connect(self.update_mapped_image)
-        self.render_button.clicked.connect(self.save_image)
+        self.render_button.clicked.connect(self.export_image)
         self.render_image_checkbox.toggled.connect(self.update_render_resolution)
         self.render_cad_checkbox.toggled.connect(self.update_render_resolution)
         self.overlay_opacity_slider.valueChanged.connect(self.change_overlay_colour)
@@ -212,7 +216,7 @@ class ImageAnalyser(CalcamGUIWindow):
         self.update_render_resolution()
 
 
-    def save_image(self):
+    def export_image(self):
 
         filename = self.get_save_filename('image')
         if filename is None:
@@ -245,7 +249,7 @@ class ImageAnalyser(CalcamGUIWindow):
             im = np.hstack((im,render_hires(self.renderer_3d,oversampling=oversampling)))
 
         im[:,:,:3] = im[:,:,2::-1]
-        cv2.imwrite(filename,im)
+        self.save_image(im,filename)
 
         if c3d is not None:
             self.renderer_3d.AddActor(c3d)
@@ -279,7 +283,7 @@ class ImageAnalyser(CalcamGUIWindow):
         visible_linewidth = 3
         visible_colour = (0,0.8,0)
         invisible_linewidth = 1
-        invisible_colour = (0.8,0,0)
+        invisible_colour = (1.,0,1.)
 
         if self.calibration is not None and self.image is not None and coords_3d is not None:
 
@@ -433,7 +437,7 @@ class ImageAnalyser(CalcamGUIWindow):
                     elif np.any(coords != self.coords_2d[i]):
                         raydata = raycast_sightlines(self.calibration,self.cadmodel,coords[0],coords[1],coords='Display',verbose=False)
                         self.update_from_3d(raydata.ray_end_coords[0,:])
-                        return
+                        break
 
 
     def on_close(self):
@@ -650,7 +654,7 @@ class ImageAnalyser(CalcamGUIWindow):
 
         self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
         self.statusbar.showMessage('Generating model octree...')
-        self.cadmodel.get_cell_locator()
+        self.cadmodel.build_octree()
         self.statusbar.clearMessage()
         self.app.restoreOverrideCursor()
         self.overlay = None
@@ -954,7 +958,7 @@ class ImageAnalyser(CalcamGUIWindow):
     def on_change_cad_features(self):
         self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
         self.statusbar.showMessage('Updating model octree...')
-        self.cadmodel.get_cell_locator()
+        self.cadmodel.build_octree()
         self.statusbar.clearMessage()
         self.app.restoreOverrideCursor()
         if self.overlay_checkbox.isChecked():
