@@ -1,5 +1,5 @@
 '''
-* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
+* Copyright 2015-2025 European Atomic Energy Community (EURATOM)
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
   will be approved by the European Commission - subsequent
@@ -24,6 +24,7 @@ import cv2
 from .core import *
 from .vtkinteractorstyles import CalcamInteractorStyle3D
 from ..calibration import Calibration
+from copy import deepcopy
 
 # View designer window.
 # This allows creation of FitResults objects for a 'virtual' camera.
@@ -78,7 +79,8 @@ class VirtualCalib(CalcamGUIWindow):
         self.x_pixels_box.valueChanged.connect(self.update_intrinsics)
         self.y_pixels_box.valueChanged.connect(self.update_intrinsics)
         self.focal_length_box.valueChanged.connect(self.update_intrinsics)
-        self.fov_box.valueChanged.connect(self.update_intrinsics)
+        self.hfov_box.valueChanged.connect(self.update_intrinsics)
+        self.vfov_box.valueChanged.connect(self.update_intrinsics)
         self.load_chessboard_button.clicked.connect(self.update_chessboard_intrinsics)
         self.load_intrinsics_button.clicked.connect(self.update_intrinsics)
         self.load_extrinsics_button.clicked.connect(self.load_viewport_calib)
@@ -98,9 +100,10 @@ class VirtualCalib(CalcamGUIWindow):
         self.intrinsics_calib = None
 
         self.fov_enabled = False
-        self.viewdir_at_cc = True
+        self.viewdir_at = 'cc'
 
         self.extrinsics_src = None
+        self.chessboard_pointpairs = None
 
         self.calibration = Calibration(cal_type='virtual')
 
@@ -180,16 +183,25 @@ class VirtualCalib(CalcamGUIWindow):
             self.load_intrinsics_button.setEnabled(False)
             self.pinhole_params_box.show()
 
-            if self.sender() is self.fov_box:
+            if self.sender() is self.vfov_box:
                 self.focal_length_box.blockSignals(True)
-                fl = self.y_pixels_box.value() * self.pixel_size_box.value() * 1e-3 / (2*np.tan(self.fov_box.value()/360 * np.pi))
+                fl = self.y_pixels_box.value() * self.pixel_size_box.value() * 1e-3 / (2*np.tan(self.vfov_box.value()/360 * np.pi))
                 self.focal_length_box.setValue(fl)
                 self.focal_length_box.blockSignals(False)
-            elif self.sender() is self.focal_length_box:
-                self.fov_box.blockSignals(True)
-                fov = 360*np.arctan(self.y_pixels_box.value() * self.pixel_size_box.value() * 1e-3 / (2*self.focal_length_box.value())) / np.pi
-                self.fov_box.setValue(fov)
-                self.fov_box.blockSignals(False)
+            elif self.sender() is self.hfov_box:
+                self.focal_length_box.blockSignals(True)
+                fl = self.x_pixels_box.value() * self.pixel_size_box.value() * 1e-3 / (2*np.tan(self.hfov_box.value()/360 * np.pi))
+                self.focal_length_box.setValue(fl)
+                self.focal_length_box.blockSignals(False)
+
+            self.hfov_box.blockSignals(True)
+            self.vfov_box.blockSignals(True)
+            hfov = 360*np.arctan(self.x_pixels_box.value() * self.pixel_size_box.value() * 1e-3 / (2*self.focal_length_box.value())) / np.pi
+            vfov = 360 * np.arctan(self.y_pixels_box.value() * self.pixel_size_box.value() * 1e-3 / (2 * self.focal_length_box.value())) / np.pi
+            self.hfov_box.setValue(hfov)
+            self.vfov_box.setValue(vfov)
+            self.hfov_box.blockSignals(False)
+            self.vfov_box.blockSignals(False)
 
             nx = self.x_pixels_box.value()
             ny = self.y_pixels_box.value()
@@ -212,6 +224,7 @@ class VirtualCalib(CalcamGUIWindow):
                     self.current_intrinsics_combobox.click()
                 return
             else:
+                self.calibration.pixel_size = None
                 self.calibration.set_chessboard_intrinsics(self.chessboard_fit,self.chessboard_pointpairs,self.chessboard_src)
                 self.current_intrinsics_combobox = self.chessboard_intrinsics
 
@@ -358,14 +371,14 @@ class VirtualCalib(CalcamGUIWindow):
         # Load the intrinsics
         if opened_calib.intrinsics_type == 'pinhole':
 
+            im_size = opened_calib.geometry.get_display_shape()
+
             if opened_calib.pixel_size is not None:
                 fl = opened_calib.view_models[0].cam_matrix[0,0] * opened_calib.pixel_size  * 1000
                 self.pixel_size_box.setValue(opened_calib.pixel_size * 1e6)
             else:
                 fl = opened_calib.view_models[0].cam_matrix[0,0] * self.pixel_size_box.value()  / 1000
 
-            im_size = opened_calib.geometry.get_display_shape()
-
             self.x_pixels_box.setValue(im_size[0])
             self.y_pixels_box.setValue(im_size[1])
 
@@ -379,8 +392,10 @@ class VirtualCalib(CalcamGUIWindow):
             self.chessboard_intrinsics.setChecked(True)
 
         elif opened_calib.intrinsics_type == 'calibration':
-            self.intrinsics_calib = self.calibration
+            self.intrinsics_calib = deepcopy(self.calibration)
             self.calcam_intrinsics.setChecked(True)
+            original_fname = opened_calib.history['intrinsics'][1].split('"')[1]
+            self.calcam_intrinsics_fname.setText(original_fname)
 
         self.update_intrinsics()
 
@@ -419,6 +434,7 @@ class VirtualCalib(CalcamGUIWindow):
         self.chessboard_fit = None
         self.intrinsics_calib = None
         self.pinhole_intrinsics.setChecked(True)
+        self.update_intrinsics()
 
         self.unsaved_changes = False
         self.refresh_3d()
\ No newline at end of file
