@@ -1,5 +1,5 @@
 '''
-* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
+* Copyright 2015-2025 European Atomic Energy Community (EURATOM)
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
   will be approved by the European Commission - subsequent
@@ -42,7 +42,7 @@ import time
 
 class CalcamInteractorStyle3D(vtk.vtkInteractorStyleTerrain):
  
-    def __init__(self,parent=None,viewport_callback=None,resize_callback=None,newpick_callback=None,cursor_move_callback=None,focus_changed_callback=None,refresh_callback=None,pre_move_callback=None):
+    def __init__(self,parent=None,viewport_callback=None,resize_callback=None,newpick_callback=None,cursor_move_callback=None,focus_changed_callback=None,refresh_callback=None,pre_move_callback=None,save_coords_callback=None):
 
         # Set callbacks for all the mouse controls
         self.AddObserver("LeftButtonPressEvent",self.on_left_click)
@@ -63,6 +63,7 @@ class CalcamInteractorStyle3D(vtk.vtkInteractorStyleTerrain):
         self.refresh_callback = refresh_callback
         self.focus_changed_callback = focus_changed_callback
         self.pre_move_callback = pre_move_callback
+        self.save_coords_callback = save_coords_callback
         self.image_actor = None
         self.image_resizer = None
         self.force_aspect = None
@@ -154,7 +155,11 @@ class CalcamInteractorStyle3D(vtk.vtkInteractorStyleTerrain):
         """
 
         direction = None
-        key = self.interactor.GetKeySym().upper()
+
+        try:
+            key = self.interactor.GetKeySym().upper()
+        except AttributeError:
+            return
 
         # -------------------- WSAD WALKING --------------------
         if key in ['W','UP']:
@@ -166,6 +171,10 @@ class CalcamInteractorStyle3D(vtk.vtkInteractorStyleTerrain):
         elif key in ['D','RIGHT']:
             direction = 'right'
 
+        elif key == 'C':
+            if self.save_coords_callback is not None:
+                self.save_coords_callback()
+
         if direction is not None:
             current_fp = self.camera.GetFocalPoint()
             view_vect = np.array(current_fp) - np.array(self.camera.GetPosition())
@@ -483,15 +492,15 @@ class CalcamInteractorStyle3D(vtk.vtkInteractorStyleTerrain):
 
             if cid == self.focus_cursor:
                 colour = (0,0.8,0)
-                linewidth = 3
-                size = 0.025
+                linewidth = 4
+                size = 0.03
             else:
                 if cursor['colour'] is not None:
                     colour = cursor['colour']
                 else:
-                    colour = (0.8,0,0)
+                    colour = (1.,0,1.)
                 linewidth = 2
-                size = 0.0125
+                size = 0.02
 
             # Cursor size scales with camera FOV to maintain size on screen.
             size = size * (self.camera.GetViewAngle()/75)
@@ -573,11 +582,15 @@ class CalcamInteractorStyle3D(vtk.vtkInteractorStyleTerrain):
             # Camera view wider than VTK panel
             if self.force_aspect <= vtk_aspect:
                 h = self.force_aspect / vtk_aspect
+                self.renderer.Clear()
+                self.vtkwindow.Render()
                 self.renderer.SetViewport([0.,0.5-h/2.,1.0,0.5+h/2.])
 
             # Camera view taller than VTK panel
             elif self.force_aspect > vtk_aspect:
                 w = vtk_aspect / self.force_aspect
+                self.renderer.Clear()
+                self.vtkwindow.Render()
                 self.renderer.SetViewport([0.5-w/2.,0.,0.5+w/2.,1.])
 
         # Sizing of the legend
@@ -619,6 +632,7 @@ class CalcamInteractorStyle3D(vtk.vtkInteractorStyleTerrain):
 
         if self.legend is not None:
             self.renderer.RemoveActor(self.legend)
+            self.legend = None
 
         if len(legend_items) > 0:
 
@@ -1158,7 +1172,7 @@ class CalcamInteractorStyle2D(vtk.vtkInteractorStyleTerrain):
                         linewidth = 3
                         size = self.cursor_size * camscale
                     else:
-                        colour = (0.8,0,0)
+                        colour = (1.,0,1.)
                         linewidth = 2
                         size = 0.5*self.cursor_size* camscale
 
