@@ -1,5 +1,5 @@
 '''
-* Copyright 2015-2021 European Atomic Energy Community (EURATOM)
+* Copyright 2015-2025 European Atomic Energy Community (EURATOM)
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
   will be approved by the European Commission - subsequent
@@ -91,7 +91,7 @@ class ZipSaveFile():
         # that we have the necessary permissions.
         if 'w' in self.mode:
             if os.path.isfile(self.filename):
-                if not os.access(self.filename,os.W_OK):
+                if self.is_readonly():
                     raise IOError('Write permission denied on {:s}'.format(self.filename))
             elif os.path.isdir(self.filename):
                 raise IOError('Specified filename already exists but is a directory!')
@@ -119,6 +119,16 @@ class ZipSaveFile():
                     else:
                         loadlist = [name for name in zf.namelist() if not name.startswith('.large/')]
 
+                    # Check there is enough disk space and raise appropriate exception if not
+                    size_to_load = 0
+                    for fname in loadlist:
+                        size_to_load += zf.getinfo(fname).file_size
+
+                    _, _, total_avail = shutil.disk_usage(self.tempdir)
+
+                    if total_avail < size_to_load:
+                        raise IOError('Not enough free space in {:s} for temporary files ({:.0f} MiB required, {:.0f} MiB available).'.format(os.path.split(self.tempdir)[0],size_to_load/1024**2,total_avail/1024**2))
+
                     zf.extractall(self.tempdir,members=loadlist)
             
             except:
@@ -361,4 +371,4 @@ class ZipSaveFile():
     # before it's garbage collected or whatever, make sure the 
     # temp files get cleaned up.
     def __del__(self):
-        self.close()
\ No newline at end of file
+        self.close()
