@@ -1,5 +1,5 @@
 '''
-* Copyright 2015-2019 European Atomic Energy Community (EURATOM)
+* Copyright 2015-2025 European Atomic Energy Community (EURATOM)
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
 will be approved by the European Commission - subsequent
@@ -55,7 +55,10 @@ def filter_points(ref_points, new_points, n_points=50, err_limit=10):
         np.ndarray  : Nx2 array of filtered x,y coordinates in the reference image
         np.ndarray  : Nx2 array of filtered x,y coordinates in the moved image
     """
-    transform = np.matrix(cv2.cv2.estimateAffinePartial2D(new_points, ref_points)[0])
+    try:
+        transform = np.matrix(cv2.estimateAffinePartial2D(new_points, ref_points)[0])
+    except AttributeError:
+        transform = np.matrix(cv2.cv2.estimateAffinePartial2D(new_points, ref_points)[0])
 
     if transform[0,0] is None:
         return np.array([]),np.array([])
@@ -165,7 +168,10 @@ def detect_movement(ref,moved):
     if ref_points.shape[0] == 0:
         raise DetectionFailedError('Could not auto-detect a good set of matching points in these two images. Consider using manual movement correction instead.')
 
-    m = np.matrix(cv2.cv2.estimateAffinePartial2D(new_points, ref_points)[0])
+    try:
+        m = np.matrix(cv2.estimateAffinePartial2D(new_points, ref_points)[0])
+    except AttributeError:
+        m = np.matrix(cv2.cv2.estimateAffinePartial2D(new_points, ref_points)[0])
 
     if m[0, 0] is None:
         raise DetectionFailedError('Could not determine image movement automatically. Consider using manual movement correction instead.')
@@ -297,7 +303,10 @@ def phase_correlation_movement(ref,moved):
     new_points[:,0] = ref_points[:,0] + movxy[0]
     new_points[:,1] = ref_points[:,1] + movxy[1]
 
-    m = np.matrix(cv2.cv2.estimateAffinePartial2D(new_points, ref_points)[0])
+    try:
+        m = np.matrix(cv2.estimateAffinePartial2D(new_points, ref_points)[0])
+    except AttributeError:
+        m = np.matrix(cv2.cv2.estimateAffinePartial2D(new_points, ref_points)[0])
 
     mov_correction = MovementCorrection(m, ref_im.shape[:2], ref_points, new_points,'Auto-generated by {:s} on {:s} at {:s}'.format(misc.username, misc.hostname, misc.get_formatted_time()))
 
