@@ -1,5 +1,5 @@
 '''
-* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
+* Copyright 2015-2025 European Atomic Energy Community (EURATOM)
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they
   will be approved by the European Commission - subsequent
@@ -19,17 +19,10 @@
   permissions and limitations under the Licence.
 '''
 
-""" 
-Ray tracing tools for CalCam_py
-
-Written by Scott Silburn
-2015-05-17
-"""
-
-import time
 import os
 import random
 import copy
+from copy import deepcopy
 
 try:
     import vtk
@@ -37,7 +30,7 @@ except:
     vtk = None
 
 import numpy as np
-from scipy.io.netcdf import netcdf_file
+from scipy.io import netcdf_file
 
 from . import coordtransformer
 from . import misc
@@ -116,12 +109,6 @@ def raycast_sightlines(calibration,cadmodel,x=None,y=None,exclusion_radius=0.0,b
     model_extent = cadmodel.get_extent()
     model_size = model_extent[1::2] - model_extent[::2]
     max_ray_length = model_size.max() * 4
-    
-    if status_callback is not None:
-        status_callback('Getting CAD model octree...')
-
-    # Get the CAD model's octree
-    cell_locator = cadmodel.get_cell_locator()
 
     # If no pixels are specified, do the whole chip at the specified binning level.
     fullchip = False
@@ -159,7 +146,7 @@ def raycast_sightlines(calibration,cadmodel,x=None,y=None,exclusion_radius=0.0,b
     results.x[valid_mask == 0] = 0
     results.y = np.copy(y).astype('float')
     results.y[valid_mask == 0] = 0
-    results.transform = calibration.geometry
+    results.transform = deepcopy(calibration.geometry)
     if calibration.filename is not None:
         splitname = os.path.split(calibration.filename)
         results.history = 'Ray cast of calibration "{:s}" [from: {:s}] by {:s} on {:s} at {:s}'.format(splitname[1].replace('.ccc',''),splitname[0],misc.username,misc.hostname,misc.get_formatted_time())
@@ -192,19 +179,10 @@ def raycast_sightlines(calibration,cadmodel,x=None,y=None,exclusion_radius=0.0,b
 
     
     if status_callback is not None:
-        oom = np.floor( np.log(np.size(x)) / np.log(10) / 3. )
+        oom = np.floor( np.log(np.size(x)) / np.log(10) / 3. ) # Order of magnitude of number of points to do
         status_callback('Casting {:s} rays...'.format( ['{:.0f}','{:.1f}k','{:.2f}M'][int(oom)].format(np.size(x)/10**(3*oom)) ) )
 
 
-    # Some variables to give to VTK becasue of its annoying C-like interface
-    t = vtk.mutable(0)
-    pos = np.zeros(3)
-    coords_ = np.zeros(3)
-    subid = vtk.mutable(0)
-    cellid = vtk.mutable(0)
-    cell = vtk.vtkGenericCell()
-
-    last_status_update = 0.
     n_done = 0
     
     # We will do the ray casting in a random order,
@@ -223,18 +201,12 @@ def raycast_sightlines(calibration,cadmodel,x=None,y=None,exclusion_radius=0.0,b
         raystart = results.ray_start_coords[ind] + exclusion_radius * LOSDir[ind]
         rayend = results.ray_start_coords[ind] + max_ray_length * LOSDir[ind]
 
-        retval = cell_locator.IntersectWithLine(raystart,rayend,1.e-6,t,pos,coords_,subid,cellid,cell)
+        ret_vals = cadmodel.intersect_with_line(raystart,rayend,calc_normals)
 
-        if abs(retval) > 0:
-            results.ray_end_coords[ind,:] = pos[:]
+        if ret_vals[0]:
+            results.ray_end_coords[ind,:] = ret_vals[1][:]
             if calc_normals:
-                v0 = np.array(cell.GetPoints().GetPoint(2)) - np.array(cell.GetPoints().GetPoint(0))
-                v1 = np.array(cell.GetPoints().GetPoint(2)) - np.array(cell.GetPoints().GetPoint(1))
-                n = np.cross(v0,v1)
-                n = n / np.sqrt(np.sum(n**2))
-                if np.dot(LOSDir[ind],n) > 0:
-                    n = -n
-                results.model_normals[ind,:] = n
+                results.model_normals[ind,:] = ret_vals[2]
 
         elif intersecting_only:
             results.ray_end_coords[ind,:] = np.nan
@@ -243,9 +215,8 @@ def raycast_sightlines(calibration,cadmodel,x=None,y=None,exclusion_radius=0.0,b
 
         n_done = n_done + 1
         
-        if time.time() - last_status_update > 1 and status_callback is not None:
+        if status_callback is not None:
             status_callback(n_done / len(inds))
-            last_status_update = time.time()
 
     if status_callback is not None:
         status_callback(1.)
@@ -336,8 +307,8 @@ class RayData:
             f.createDimension('vdim',self.x.shape[0])
             rayhit = f.createVariable('RayEndCoords','f4',('vdim','udim','pointdim'))
             raystart = f.createVariable('RayStartCoords','f4',('vdim','udim','pointdim'))
-            x = f.createVariable('PixelXLocation','i4',('vdim','udim'))
-            y = f.createVariable('PixelYLocation','i4',('vdim','udim'))
+            x = f.createVariable('PixelXLocation','f4',('vdim','udim'))
+            y = f.createVariable('PixelYLocation','f4',('vdim','udim'))
             if self.model_normals is not None:
                 normals = f.createVariable('ModelNormals', 'f4', ('vdim', 'udim', 'pointdim'))
             
@@ -355,8 +326,8 @@ class RayData:
             if self.model_normals is not None:
                 normals = f.createVariable('ModelNormals','f4',('udim','pointdim'))
 
-            x = f.createVariable('PixelXLocation','i4',('udim',))
-            y = f.createVariable('PixelYLocation','i4',('udim',))
+            x = f.createVariable('PixelXLocation','f4',('udim',))
+            y = f.createVariable('PixelYLocation','f4',('udim',))
 
             rayhit[:,:] = self.ray_end_coords
             raystart[:,:] = self.ray_start_coords
@@ -367,12 +338,12 @@ class RayData:
         else:
             raise Exception('Cannot save RayData with >2D x and y arrays!')
 
-        binning = f.createVariable('Binning','i4',())
+        binning = f.createVariable('Binning','f4',())
 
         if self.binning is not None:
-            binning.assignValue(self.binning)
+            binning.data = np.array(self.binning,dtype='f4')
         else:
-            binning.assignValue(0)
+            binning.data = np.array(0,dtype='f4')
 
         f.createDimension('pixelsdim',2)
 
@@ -383,7 +354,7 @@ class RayData:
         offset[:] = self.transform.offset[:]
 
         pixelaspect = f.createVariable('image_original_pixel_aspect','f4',())
-        pixelaspect.assignValue(self.transform.pixel_aspectratio)
+        pixelaspect.data = np.array(self.transform.pixel_aspectratio,dtype='f4')
 
         binning.units = 'pixels'
         raystart.units = 'm'
@@ -516,10 +487,14 @@ class RayData:
             self.x = nx
             self.y = ny
 
-            if self.fullchip:
+            if isinstance(self.fullchip,bool) or self.fullchip.lower() == 'display' or (self.fullchip.lower() == 'original' and not self.transform._is_sideways()):
                 rowinds = np.squeeze(np.argwhere(np.logical_and(self.y[:,0] >= 0, self.y[:,0] < newshape[1])))
                 colinds = np.squeeze(np.argwhere(np.logical_and(self.x[rowinds,:][0,:] >= 0, self.x[rowinds,:][0,:] < newshape[0])))
                 self.crop_inds = (rowinds,colinds)
+            elif self.fullchip:
+                rowinds = np.squeeze(np.argwhere(np.logical_and(self.x[:,0] >= 0, self.x[:,0] < newshape[1])))
+                colinds = np.squeeze(np.argwhere(np.logical_and(self.y[rowinds,:][0,:] >= 0, self.y[rowinds,:][0,:] < newshape[0])))
+                self.crop_inds = (rowinds,colinds)
 
 
             elif self.x.ndim == 2:
@@ -527,7 +502,7 @@ class RayData:
                 yinds = np.argwhere(np.logical_and(self.y[xinds] >= 0, self.y[xinds] < newshape[1]))
                 self.crop_inds = (xinds,yinds)
             else:
-                raise Exception("Cropping a raydata object with strange dimensions! I don't know how to do that; help!")
+                raise NotImplementedError("Cropping a raydata object with strange dimensions! I don't know how to do that; help!")
 
             self.crop = copy.deepcopy(window)
 
@@ -559,18 +534,23 @@ class RayData:
                                       dimension added which contains the  [X,Y,Z] 3D coordinates. Otherwise the shape\
                                       is (h x w x 3) where w and h are the image width and height (in display coords).
         '''
+
+        if coords.lower() not in ['display','original']:
+            raise ValueError('Invalid coordinates argument "{:s}" - must be either "Display" or "Original".'.format(coords))
+
         if x is None and y is None:
             if self.fullchip:
-                if coords.lower() == 'display':
-                    if self.crop is None:
-                        return self.ray_start_coords
-                    else:
-                        return self.ray_start_coords[self.crop_inds[0],:,:][:,self.crop_inds[1],:]
+                if self.crop is None:
+                    raystarts = self.ray_start_coords
                 else:
-                    if self.crop is None:
-                        return self.transform.display_to_original_image(self.ray_start_coords)
-                    else:
-                        return self.transform.display_to_original_image(self.ray_start_coords[self.crop_inds[0],:,:][:,self.crop_inds[1],:])
+                    raystarts = self.ray_start_coords[self.crop_inds[0], :, :][:, self.crop_inds[1], :]
+
+                if isinstance(self.fullchip, bool) or coords.lower() == self.fullchip.lower():
+                    return raystarts
+                elif coords.lower() == 'display':
+                    return self.transform.original_to_display_image(raystarts)
+                elif coords.lower() == 'original':
+                    return self.transform.display_to_original_image(raystarts)
             else:
                 if self.crop is None:
                     return self.ray_start_coords
@@ -638,18 +618,24 @@ class RayData:
                                       is (h x w x 3) where w and h are the image width and height (in display coords).
 
         '''
+
+        if coords.lower() not in ['display','original']:
+            raise ValueError('Invalid coordinates argument "{:s}" - must be either "Display" or "Original".'.format(coords))
+
         if x is None and y is None:
             if self.fullchip:
-                if coords.lower() == 'display':
-                    if self.crop is None:
-                        return self.ray_end_coords
-                    else:
-                        return self.ray_end_coords[self.crop_inds[0],:,:][:,self.crop_inds[1],:]
+                if self.crop is None:
+                    rayends = self.ray_end_coords
                 else:
-                    if self.crop is None:
-                        return self.transform.display_to_original_image(self.ray_end_coords)
-                    else:
-                        return self.transform.display_to_original_image(self.ray_end_coords[self.crop_inds[0],:,:][:, self.crop_inds[1],:])
+                    rayends = self.ray_end_coords[self.crop_inds[0],:,:][:,self.crop_inds[1],:]
+
+                if isinstance(self.fullchip,bool) or coords.lower() == self.fullchip.lower():
+                    return rayends
+                elif coords.lower() == 'display':
+                    return self.transform.original_to_display_image(rayends)
+                elif coords.lower() == 'original':
+                    return self.transform.display_to_original_image(rayends)
+
             else:
                 if self.crop is None:
                     return self.ray_end_coords
@@ -775,6 +761,7 @@ class RayData:
 
                 return np.reshape(out,oldshape + (3,),order='F')
 
+
     def get_ray_lengths(self,x=None,y=None,im_position_tol = 1,coords='Display'):
         '''
         Get the sight-line lengths either of all casted sight-lines or at the specified image coordinates.
@@ -799,71 +786,8 @@ class RayData:
         '''
         
         # Work out ray lengths for all raytraced pixels
-        raylength = np.sqrt(np.sum( (self.ray_end_coords - self.ray_start_coords) **2,axis=-1))
-        
-        # If no x and y given, return them all
-        if x is None and y is None:
-            if self.fullchip:
-                if coords.lower() == 'display':
-                    if self.crop is None:
-                        return raylength
-                    else:
-                        return raylength[self.crop_inds[0],:][:,self.crop_inds[1]]
-                else:
-                    if self.crop is None:
-                        return self.transform.display_to_original_image(raylength)
-                    else:
-                        return self.transform.display_to_original_image(raylength[self.crop_inds[0],:][:,self.crop_inds[1]])
-            else:
-                if self.crop is None:
-                    return raylength
-                else:
-                    return raylength[self.crop_inds[0]][self.crop_inds[1]]
-        else:
-            if self.x is None or self.y is None:
-                raise Exception('This ray data does not have x and y pixel indices!')
-
-            # Otherwise, return the ones at given x and y pixel coords.
-            if np.shape(x) != np.shape(y):
-                raise ValueError('x and y arrays must be the same shape!')
-            else:
-
-                if coords.lower() == 'original':
-                    x,y = self.transform.original_to_display_coords(x,y)
-
-                orig_shape = np.shape(x)
-                x = np.reshape(x,np.size(x),order='F')
-                y = np.reshape(y,np.size(y),order='F')
-                RL = np.zeros(np.shape(x))
-
-                if not self.fullchip:
-                    raylength = raylength.flatten()
-                    xflat = self.x.flatten()
-                    yflat = self.y.flatten()
-
-                for pointno in range(x.size):
-                    if np.isnan(x[pointno]) or np.isnan(y[pointno]):
-                        RL[pointno] = np.nan
-                        continue
-
-                    if self.fullchip:
-                        xind = np.argmin(np.abs(self.x[0,:] - x[pointno]))
-                        yind = np.argmin(np.abs(self.y[:,0] - y[pointno]))
-                        deltaR = np.sqrt( (self.x[0,xind]-x[pointno])**2 + (self.y[yind,0]-y[pointno])**2)
-                        if deltaR < im_position_tol:
-                            RL[pointno] = raylength[yind,xind]
-                        else:
-                            raise Exception('No ray-traced pixel within im_position_tol of requested pixel ({:.1f},{:.1f})!'.format(x[pointno], y[pointno]))
-                    else:
-                        # This can be very slow if xflat and yflat are big arrays!
-                        deltaR = np.sqrt( (xflat - x[pointno])**2 + (yflat - y[pointno])**2 )
-                        if np.nanmin(deltaR) <= im_position_tol:
-                            RL[pointno] = raylength[np.nanargmin(deltaR)]
-                        else:
-                            raise Exception('No ray-traced pixel within im_position_tol of requested pixel!')
-
-                return np.reshape(RL,orig_shape,order='F')
-
+        ray_vectors = self.get_ray_end(x,y,im_position_tol,coords) - self.get_ray_start(x,y,im_position_tol,coords)
+        return np.sqrt(np.sum(ray_vectors**2,axis=-1))
 
 
     def get_ray_directions(self,x=None,y=None,im_position_tol=1,coords='Display'):
@@ -891,59 +815,6 @@ class RayData:
                                       (h x w x 3) where w nd h are the image width and height. Otherwise it will \
                                       be the same shape as the input x and y coordinates plus an extra dimension.
         '''
-        vectors = (self.ray_end_coords - self.ray_start_coords)
+        vectors = self.get_ray_end(x,y,im_position_tol,coords) - self.get_ray_start(x,y,im_position_tol,coords)
         lengths = np.sqrt(np.sum(vectors**2,axis=-1))
-        dirs =  vectors / np.repeat(lengths.reshape(np.shape(lengths)+(1,)),3,axis=-1)
-
-        if x is None and y is None:
-            if self.fullchip:
-                if coords.lower() == 'display':
-                    if self.crop is None:
-                        return dirs
-                    else:
-                        return dirs[self.crop_inds[0],:,:][:,self.crop_inds[1],:]
-                else:
-                    if self.crop is None:
-                        return self.transform.display_to_original_image(dirs)
-                    else:
-                        return self.transform.display_to_original_image(dirs[self.crop_inds[0],:,:][:, self.crop_inds[1],:])
-            else:
-                if self.crop is None:
-                    return dirs
-                else:
-                    return dirs[self.crop_inds[0]][self.crop_inds[1]]
-        else:
-            if self.x is None or self.y is None:
-                raise Exception('This ray data does not have x and y pixel indices!')
-            if np.shape(x) != np.shape(y):
-                raise ValueError('x and y arrays must be the same shape!')
-            else:
-
-                if coords.lower() == 'original':
-                    x,y = self.transform.original_to_display_coords(x,y)
-
-                oldshape = np.shape(x)
-                x = np.reshape(x,np.size(x),order='F')
-                y = np.reshape(y,np.size(y),order='F')
-                [dirs_X,dirs_Y,dirs_Z] = np.split(dirs,3,-1)
-                dirs_X = dirs_X.flatten()
-                dirs_Y = dirs_Y.flatten()
-                dirs_Z = dirs_Z.flatten()
-                xflat = self.x.flatten()
-                yflat = self.y.flatten()
-                Xout = np.zeros(np.shape(x))
-                Yout = np.zeros(np.shape(x))
-                Zout = np.zeros(np.shape(x))
-                for pointno in range(x.size):
-                    deltaX = xflat - x[pointno]
-                    deltaY = yflat - y[pointno]
-                    deltaR = np.sqrt(deltaX**2 + deltaY**2)
-                    if np.min(deltaR) <= im_position_tol:
-                        Xout[pointno] = dirs_X[np.argmin(deltaR)]
-                        Yout[pointno] = dirs_Y[np.argmin(deltaR)]
-                        Zout[pointno] = dirs_Z[np.argmin(deltaR)]
-                    else:
-                        raise Exception('No ray-traced pixel within im_position_tol of requested pixel!')
-                out = np.hstack([Xout,Yout,Zout])
-
-                return np.reshape(out,oldshape + (3,),order='F')
+        return  vectors / np.expand_dims(lengths,-1)
\ No newline at end of file
